# -*- coding: utf-8 -*-
from flask import current_app as app
from flask.ext.restful import fields, marshal

from shiva.converter import get_converter
from shiva.media import get_mimetypes


class InstanceURI(fields.String):
    def __init__(self, base_uri):
        server_uri = app.config.get('SERVER_URI') or ''
        self.base_uri = '/'.join((server_uri, base_uri))

    def output(self, key, obj):
        return '/'.join((self.base_uri, str(obj.pk)))


class TrackFiles(fields.Raw):
    """
    Returns a list of files, one for each available mediatype, for a given
    track.

    """

    def output(self, key, track):
        ConverterClass = get_converter()
        paths = {}

        for mimetype in get_mimetypes():
            converter = ConverterClass(track, mimetype)
            paths[str(mimetype)] = converter.get_uri()

        return paths


class ManyToManyField(fields.Raw):
    def __init__(self, foreign_obj, nested):
        self.foreign_obj = foreign_obj
        self.nested = nested

        super(ManyToManyField, self).__init__()

    def output(self, key, obj):
        items = list()
        for item in getattr(obj, key):
            items.append(marshal(item, self.nested))

        return items


class ForeignKeyField(fields.Raw):
    def __init__(self, foreign_obj, nested):
        self.foreign_obj = foreign_obj
        self.nested = nested

        super(ForeignKeyField, self).__init__()

    def output(self, key, obj):
        _id = getattr(obj, '%s_pk' % key)
        if not _id:
            return None

        obj = self.foreign_obj.query.get(_id)

        return marshal(obj, self.nested)


class Boolean(fields.Raw):
    def output(self, key, obj):
        return bool(super(Boolean, self).output(key, obj))



import re
import urllib2

import requests

from shiva.lyrics import LyricScraper
from shiva.utils import get_logger

log = get_logger()


class AZLyrics(LyricScraper):
    """
    """

    def __init__(self, artist, title):
        self.artist = artist
        self.title = title
        self.lyrics = None
        self.source = None

        self.search_url = 'http://search.azlyrics.com/search.php?q=%s'
        self.lyric_url_re = re.compile(r'http://www\.azlyrics\.com/lyrics/'
                                       r'[a-z0-9]+/[a-z0-9]+\.html')
        self.lyric_re = re.compile(r'<!-- start of lyrics -->(.*)'
                                   r'<!-- end of lyrics -->', re.M + re.S)
        self.title_re = re.compile(r'<title>(?P<artist>.*?) LYRICS - '
                                   r'(?P<title>.*?)</title>')

    def fetch(self):
        self.search()
        if not self.source:
            return None

        response = requests.get(self.source)
        self.html = response.text

        if not self.check():
            return False

        log.info('[FOUND] %s' % self.source)
        lyrics = self.lyric_re.findall(self.html)[0]
        lyrics = re.sub(r'<br[ /]?>', '\r', lyrics)
        lyrics = re.sub(r'<.*?>', '', lyrics)

        self.lyrics = lyrics.strip()

        return True

    def search(self):
        query = urllib2.quote('%s %s' % (self.artist, self.title))
        log.info('[SEARCH] %s' % (self.search_url % query))
        response = requests.get(self.search_url % query)
        results = self.lyric_url_re.findall(response.text)

        if results:
            self.source = results[0]

    def check(self):
        match = self.title_re.search(self.html)

        if match.group('artist').lower() != self.artist.lower():
            return False

        if match.group('title').lower() != self.title.lower():
            return False

        return True



# -*- coding: utf-8 -*-
from nose import tools as nose
import os
import tempfile
import unittest

from shiva import app as shiva


class StatusCodesTestCase(unittest.TestCase):

    def setUp(self):
        self.db_fd, self.db_path = tempfile.mkstemp()
        db_uri = 'sqlite:///%s' % self.db_path
        shiva.app.config['SQLALCHEMY_DATABASE_URI'] = db_uri
        shiva.app.config['TESTING'] = True
        shiva.db.create_all()
        self.app = shiva.app.test_client()

    def tearDown(self):
        os.close(self.db_fd)
        os.unlink(self.db_path)

    def test_root(self):
        rv = self.app.get('/')
        nose.eq_(rv.status_code, 404)

    def test_artists(self):
        rv = self.app.get('/artists')
        nose.eq_(rv.status_code, 200)

    def test_artist_404(self):
        rv = self.app.get('/artist/1')
        nose.eq_(rv.status_code, 404)

    def test_albums(self):
        rv = self.app.get('/albums')
        nose.eq_(rv.status_code, 200)

    def test_album_404(self):
        rv = self.app.get('/album/1')
        nose.eq_(rv.status_code, 404)

    def test_tracks(self):
        rv = self.app.get('/tracks')
        nose.eq_(rv.status_code, 200)

    def test_track_404(self):
        rv = self.app.get('/track/1')
        nose.eq_(rv.status_code, 404)



# -*- coding: utf-8 -*-
from nose import tools as nose
import unittest
from mock import Mock

from shiva import app


class AppTestCase(unittest.TestCase):

    def setUp(self):
        app.app.run = Mock()

    def test_main(self):
        app.main()
        nose.assert_true(app.app.run.called)



from flask import current_app as app, Response
from flask.ext import restful

from shiva.decorators import allow_origins


class Resource(restful.Resource):
    def __new__(cls, *args, **kwargs):
        if app.config.get('CORS_ENABLED') is True:
            # Applies to all inherited resources
            cls.method_decorators = [allow_origins]

        return super(Resource, cls).__new__(cls, *args, **kwargs)

    # Without this the shiva.decorator.allow_origins method won't get called
    # when issuing an OPTIONS request.
    def options(self, *args, **kwargs):
        return JSONResponse()


class JSONResponse(Response):
    """
    A subclass of flask.Response that sets the Content-Type header by default
    to "application/json".

    """

    def __init__(self, status=200, **kwargs):
        params = {
            'headers': [],
            'mimetype': 'application/json',
            'response': '',
            'status': status,
        }
        params.update(kwargs)

        super(JSONResponse, self).__init__(**params)



# -*- coding: utf-8 -*-
import os
import subprocess
import urllib

from flask import current_app as app

from shiva.exceptions import InvalidMimeTypeError
from shiva.media import MimeType


def get_converter():
    ConverterClass = app.config.get('CONVERTER_CLASS', Converter)

    return ConverterClass


class Converter(object):
    """
    Class responsible for format conversion. Receives the path to a track (as a
    string) and decides where to save the converted files. It also decides
    which external software to use for conversion.

    """

    CONVERSION_URI = '/track/%s/convert?%s'

    def __init__(self, track, mimetype):
        self.track = track
        self.mimetypes = self.get_mimetypes()
        self.set_mimetype(mimetype)

    def get_mimetypes(self):
        """
        Returns the list of mimetypes that the system supports, provided by the
        MIMETYPES setting.

        """

        return app.config.get('MIMETYPES', [])

    def set_mimetype(self, mimetype):
        """
        Sets the mimetype or raises an InvalidMimeTypeError exception if it's
        not found in the settings.
        """

        mimetype_object = None
        if issubclass(mimetype.__class__, MimeType):
            mimetype_object = mimetype
        else:
            for _mime in self.mimetypes:
                if hasattr(_mime, 'matches') and callable(_mime.matches):
                    if _mime.matches(mimetype):
                        mimetype_object = _mime

        if mimetype_object:
            self.fullpath = None
            self.uri = None
            self.mimetype = mimetype_object
        else:
            raise InvalidMimeTypeError(mimetype)

    def get_dest_directory(self, mime=None):
        """
        Retrieves the path on which a track's duplicates, i.e. that same track
        converted to other formats, will be stored.

        In this case they will be stored along the original track, in the same
        directory.

        """

        return os.path.dirname(self.track.path)

    def get_dest_filename(self):
        filename = os.path.basename(self.track.path)
        filename = '.'.join(filename.split('.')[0:-1])

        return '.'.join((filename, self.mimetype.extension))

    def get_dest_fullpath(self):
        if self.fullpath:
            return self.fullpath

        directory = self.get_dest_directory()
        filename = self.get_dest_filename()
        self.fullpath = os.path.join(directory, filename)

        return self.fullpath

    def get_conversion_uri(self):
        mimetype = urllib.urlencode({'mimetype': str(self.mimetype)})
        uri = self.CONVERSION_URI % (self.track.pk, mimetype)

        return ''.join((app.config.get('SERVER_URI') or '', uri))

    def get_uri(self):
        if self.converted_file_exists():
            self.uri = self.get_file_uri()
        else:
            self.uri = self.get_conversion_uri()

        return self.uri

    def get_file_uri(self):
        for mdir in app.config['MEDIA_DIRS']:
            return mdir.urlize(self.get_dest_fullpath())

    def convert(self):
        path = self.get_dest_fullpath()
        if self.converted_file_exists():
            return path

        # Don't do app.config.get('FFMPEG_PATH', 'ffmpeg'). That will cause an
        # error when FFMPEG_PATH is set to None or empty string.
        ffmpeg = app.config.get('FFMPEG_PATH') or 'ffmpeg'
        cmd = [ffmpeg, '-i', self.track.path, '-aq', '60', '-acodec',
               self.mimetype.acodec, path]

        proc = subprocess.call(cmd)

        return path

    def converted_file_exists(self):
        return os.path.exists(self.get_dest_fullpath())



# -*- coding: utf-8 -*-
from shiva.utils import ignored, get_logger

log = get_logger()


class LastFM(object):
    def __init__(self, api_key, use_cache=True):
        """
        Wrapper around LastFM's library, pylast. Simplifies handling and adds
        caching.

        Cache's schema:
            self.cache[artist.name]
            self.cache[artist.name]['object']
            self.cache[artist.name]['albums']
            self.cache[artist.name]['albums'][album.name]
        """
        import pylast

        self.pylast = pylast
        self.use_cache = use_cache
        self.lib = self.pylast.LastFMNetwork(api_key=api_key)
        self.cache = {}

    def get_artist(self, name):
        artist = self.cache.get(name, {}).get('object')

        if not artist:
            with ignored(Exception, print_traceback=True):
                log.debug('[ Last.FM ] Retrieving artist "%s"' % name)
                artist = self.lib.get_artist(name)
            if artist and self.use_cache:
                self.cache.update({
                    name: {
                        'object': artist
                    }
                })

        return artist

    def get_artist_image(self, name):
        image = None
        if self.use_lastfm:
            log.debug('[ Last.FM ] Retrieving artist image for "%s"' % name)
            with ignored(Exception, print_traceback=True):
                image = self.get_artist(name).get_cover_image()

    def get_album(self, name, artist_name):
        album = self.cache.get(artist_name, {}).get('albums', {}).get(name)

        if not album:
            artist = self.get_artist(artist_name)
            log.debug('[ Last.FM ] Retrieving album "%s" by "%s"' % (
                      name, artist.name))
            with ignored(Exception, print_traceback=True):
                album = self.lib.get_album(artist, name)

            if album and self.use_cache:
                self.cache.update({
                    artist.name: {
                        'albums': {
                            album.name: album
                        }
                    }
                })

        return album

    def get_release_date(self, album_name, artist_name):
        album = self.get_album(album_name, artist_name)
        if not album:
            return None

        log.debug('[ Last.FM ] album "%s" by "%s" release date' % (
                  name, artist.name))
        with ignored(Exception, print_traceback=True):
            rdate = album.get_release_date()

        if not rdate:
            return None

        return datetime.strptime(rdate, '%d %b %Y, %H:%M')

    def get_album_cover(self, album_name, artist_name):
        cover = None
        album = self.get_album(album_name, artist_name)

        if not album:
            return None

        log.debug('[ Last.FM ] Retrieving album "%s" by "%s" cover image' % (
                  name, artist.name))
        with ignored(Exception, print_traceback=True):
            cover = album.get_cover_image(size=self.lib.COVER_EXTRA_LARGE)

        return cover

    def clear_cache(self):
        self.cache = {}



# -*- coding: utf-8 -*-
from nose import tools as nose
import unittest

from shiva import exceptions as exc


class ExceptionsTestCase(unittest.TestCase):

    def test_invalid_mimetype_error(self):
        error = exc.InvalidMimeTypeError('audio/mp3')
        nose.eq_(error.__str__(), "Invalid mimetype 'audio/mp3'")

    def test_no_config_found_error(self):
        error = exc.NoConfigFoundError()
        nose.assert_not_equal(error.__str__(), '')



from functools import wraps
from flask import g, request, make_response
from flask import current_app as app


def allow_origins(func=None, custom_origins=None):
    """
    Add headers for Cross-origin resource sharing based on
    `CORS_ALLOWED_ORIGINS` in config, or parameters passed to the decorator.
    `CORS_ALLOWED_ORIGINS` can be a list of allowed origins or `"*"` to allow
    all origins.

    """

    def wrapped(func):
        def _get_origin(allowed_origins, origin):
            """ Helper method to discover the proper value for
            Access-Control-Allow-Origin to use.

            If the allowed origin is a string it will check if it's '*'
            wildcard or an actual domain. When a tuple or list is given
            instead, it will look for the current domain in the list. If any of
            the checks fail it will return False.

            """

            if type(allowed_origins) in (str, unicode):
                if allowed_origins == '*' or allowed_origins == origin:
                    return allowed_origins

            elif type(allowed_origins) in (list, tuple):
                if origin in allowed_origins:
                    return origin

            return False

        @wraps(func)
        def decorated(*args, **kwargs):
            origin = request.headers.get('Origin')

            # `app.config.get('CORS_ALLOWED_ORIGINS', [])` should really be the
            # default option in `def allow_origins` for `custom_origins` but
            # that would use `app` outside of the application context
            allowed_origins = custom_origins or \
                app.config.get('CORS_ALLOWED_ORIGINS', [])

            # Actual headers are added in `after_request`
            g.cors = _get_origin(allowed_origins, origin)

            return func(*args, **kwargs)

        return decorated

    if func:
        return wrapped(func)

    return wrapped



# -*- coding: utf-8 -*-
from mock import Mock
try:
    import unittest2 as unittest
except ImportError:
    import unittest
import os

from shiva.indexer.main import Indexer


class IndexerTestCase(unittest.TestCase):

    def setUp(self):
        self.config = {
            'MEDIA_DIRS': [],
            'VALID_FILE_EXTENSIONS': ('mp3',),
        }
        self.indexer = Indexer(self.config)

        os.path.isfile = Mock()

    def test_track_detection(self):
        self.indexer.file_path = 'valid_track.mp3'

        self.assertTrue(self.indexer.is_track())

    def test_extension_detection(self):
        self.indexer.file_path = 'valid_track.mp3'

        self.assertEqual(self.indexer.get_extension(), 'mp3')

    def test_runs(self):
        self.assertIsNone(self.indexer.run())



# -*- coding: utf-8 -*-
from datetime import datetime
import os

from flask.ext.sqlalchemy import SQLAlchemy
from sqlalchemy.exc import OperationalError
from sqlalchemy.sql.expression import func

from shiva.utils import slugify, MetadataManager

db = SQLAlchemy()

__all__ = ('db', 'Artist', 'Album', 'Track', 'LyricsCache')


def random_row(model):
    """Retrieves a random row for the given model."""

    try:
        # PostgreSQL, SQLite
        instance = model.query.order_by(func.random()).limit(1).first()
    except OperationalError:
        # MySQL
        instance = model.query.order_by(func.rand()).limit(1).first()

    return instance


class Artist(db.Model):
    """
    """

    __tablename__ = 'artists'

    pk = db.Column(db.Integer, primary_key=True)
    # TODO: Update the files' Metadata when changing this info.
    name = db.Column(db.String(128), nullable=False)
    slug = db.Column(db.String(128), nullable=False)
    image = db.Column(db.String(256))
    events = db.Column(db.String(256))
    date_added = db.Column(db.Date(), nullable=False)

    tracks = db.relationship('Track', backref='artist', lazy='dynamic')

    def __init__(self, *args, **kwargs):
        if 'date_added' not in kwargs:
            kwargs['date_added'] = datetime.today()

        super(Artist, self).__init__(*args, **kwargs)

    @classmethod
    def random(cls):
        return random_row(cls)

    def __setattr__(self, attr, value):
        if attr == 'name':
            super(Artist, self).__setattr__('slug', slugify(value))

        super(Artist, self).__setattr__(attr, value)

    def __repr__(self):
        return '<Artist (%s)>' % self.name


artists = db.Table('albumartists',
    db.Column('artist_pk', db.Integer, db.ForeignKey('artists.pk')),
    db.Column('album_pk', db.Integer, db.ForeignKey('albums.pk'))
)


class Album(db.Model):
    """
    """

    __tablename__ = 'albums'

    pk = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(128), nullable=False)
    slug = db.Column(db.String(128), nullable=False)
    year = db.Column(db.Integer)
    cover = db.Column(db.String(256))
    date_added = db.Column(db.Date(), nullable=False)

    tracks = db.relationship('Track', backref='album', lazy='dynamic')

    artists = db.relationship('Artist', secondary=artists,
                              backref=db.backref('albums', lazy='dynamic'))

    def __init__(self, *args, **kwargs):
        if 'date_added' not in kwargs:
            kwargs['date_added'] = datetime.today()

        super(Album, self).__init__(*args, **kwargs)

    @classmethod
    def random(cls):
        return random_row(cls)

    def __setattr__(self, attr, value):
        if attr == 'name':
            super(Album, self).__setattr__('slug', slugify(value))

        super(Album, self).__setattr__(attr, value)

    def __repr__(self):
        return '<Album (%s)>' % self.name


class Track(db.Model):
    """Track model."""

    __tablename__ = 'tracks'

    pk = db.Column(db.Integer, primary_key=True)
    path = db.Column(db.Unicode(256), unique=True, nullable=False)
    title = db.Column(db.String(128))
    slug = db.Column(db.String(128))
    bitrate = db.Column(db.Integer)
    file_size = db.Column(db.Integer)
    # TODO could be float if number weren't converted to an int in metadata
    # manager
    length = db.Column(db.Integer)
    # TODO number should probably be renamed to track or track_number
    number = db.Column(db.Integer)
    date_added = db.Column(db.Date(), nullable=False)

    lyrics = db.relationship('LyricsCache', backref='track', uselist=False)

    album_pk = db.Column(db.Integer, db.ForeignKey('albums.pk'), nullable=True)
    artist_pk = db.Column(db.Integer, db.ForeignKey('artists.pk'),
                          nullable=True)

    def __init__(self, path, *args, **kwargs):
        if not isinstance(path, (basestring, file)):
            raise ValueError('Invalid parameter for Track. Path or File '
                             'expected, got %s' % type(path))

        _path = path
        if isinstance(path, file):
            _path = path.name

        no_metadata = False
        if 'no_metadata' in kwargs:
            no_metadata = kwargs.get('no_metadata', False)
            del(kwargs['no_metadata'])

        self._meta = None
        self.set_path(_path, no_metadata=no_metadata)

        if 'date_added' not in kwargs:
            kwargs['date_added'] = datetime.today()

        super(Track, self).__init__(*args, **kwargs)

    @classmethod
    def random(cls):
        return random_row(cls)

    def __setattr__(self, attr, value):
        if attr == 'title':
            super(Track, self).__setattr__('slug', slugify(value))

        super(Track, self).__setattr__(attr, value)

    def get_path(self):
        if self.path:
            return self.path.encode('utf-8')

        return None

    def set_path(self, path, no_metadata=False):
        if path != self.get_path():
            self.path = path
            if no_metadata:
                return None

            if os.path.exists(self.get_path()):
                meta = self.get_metadata_reader()
                self.file_size = meta.filesize
                self.bitrate = meta.bitrate
                self.length = meta.length
                self.number = meta.track_number
                self.title = meta.title

    def get_metadata_reader(self):
        """Return a MetadataManager object."""
        if not getattr(self, '_meta', None):
            self._meta = MetadataManager(self.get_path())

        return self._meta

    def __repr__(self):
        return "<Track ('%s')>" % self.title


class LyricsCache(db.Model):
    """
    """

    __tablename__ = 'lyricscache'

    pk = db.Column(db.Integer, primary_key=True)
    text = db.Column(db.Text)
    source = db.Column(db.String(256))

    track_pk = db.Column(db.Integer, db.ForeignKey('tracks.pk'),
                         nullable=False)

    def __repr__(self):
        return "<LyricsCache ('%s')>" % self.track.title



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

import shiva

setup(
    name='shiva',
    version=shiva.__version__,
    description='A RESTful API to your music collection',
    long_description=shiva.__doc__,
    author=u'Alvaro Mouriño',
    author_email='alvaro@mourino.net',
    url=shiva.__url__,
    download_url='https://codeload.github.com/tooxie/shiva-server/zip/master',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2 :: Only',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python',
        'Topic :: Communications :: File Sharing',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Server',
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Multimedia',
    ],
    license='MIT',
    platforms=['OS-independent', 'Any'],
    package_dir={'': '.'},
    packages=find_packages('.'),
    install_requires=[
        'docopt==0.6.1',
        'Flask-Restful==0.2.3',
        'Flask-SQLAlchemy==0.16',
        'Flask==0.10',
        'lxml==3.1beta1',
        'mutagen==1.21',
        'pyLast==0.5.11',
        'python-dateutil==2.1',
        'python-slugify==0.0.3',
        'requests==1.0.4',
    ],
    entry_points={
        'console_scripts': [
            'shiva-server = shiva.app:main',
            'shiva-indexer = shiva.indexer:main',
            'shiva-fileserver = shiva.fileserver:main',
        ]
    }
)



# -*- coding: utf-8 -*-
from datetime import datetime
import urllib2
import traceback

from flask import request, current_app as app, g
from flask.ext.restful import abort, fields, marshal
from werkzeug.exceptions import NotFound
import requests

from shiva import get_version, get_contributors
from shiva.converter import get_converter
from shiva.exceptions import InvalidMimeTypeError
from shiva.fields import (Boolean, ForeignKeyField, InstanceURI,
                          ManyToManyField, TrackFiles)
from shiva.http import Resource, JSONResponse
from shiva.lyrics import get_lyrics
from shiva.mocks import ShowModel
from shiva.models import Artist, Album, Track, LyricsCache
from shiva.utils import get_logger

log = get_logger()


def full_tree():
    """ Checks the GET parameters to see if a full tree was requested. """

    arg = request.args.get('fulltree')

    return (arg and arg not in ('false', '0'))


def paginate(queryset):
    """
    Function that receives a queryset and paginates it based on the GET
    parameters.

    """

    try:
        page_size = int(request.args.get('page_size', 0))
    except ValueError:
        page_size = 0

    try:
        page_number = int(request.args.get('page', 0))
    except ValueError:
        page_number = 0

    if not page_size or not page_number:
        return queryset

    total = queryset.count()

    limit = page_size
    offset = page_size * (page_number - 1)

    return queryset.limit(limit).offset(offset)


class ArtistResource(Resource):
    """ The resource resposible for artists. """

    def get_resource_fields(self):
        return {
            'id': fields.Integer(attribute='pk'),
            'name': fields.String,
            'slug': fields.String,
            'uri': InstanceURI('artist'),
            'image': fields.String(default=app.config['DEFAULT_ARTIST_IMAGE']),
            'events_uri': fields.String(attribute='events'),
        }

    def get(self, artist_id=None, artist_slug=None):
        if not artist_id and not artist_slug:
            return list(self.get_all())

        if not artist_id and artist_slug:
            artist = self.get_by_slug(artist_slug)
        else:
            artist = self.get_one(artist_id)

        if full_tree():
            return self.get_full_tree(artist)

        return marshal(artist, self.get_resource_fields())

    def get_all(self):
        for artist in paginate(Artist.query.order_by(Artist.name)):
            yield marshal(artist, self.get_resource_fields())

    def get_one(self, artist_id):
        artist = Artist.query.get(artist_id)

        if not artist:
            abort(404)

        return artist

    def get_by_slug(self, artist_slug):
        artist = Artist.query.filter_by(slug=artist_slug).first()

        if not artist:
            abort(404)

        return artist

    def get_full_tree(self, artist):
        _artist = marshal(artist, self.get_resource_fields())
        _artist['albums'] = []

        albums = AlbumResource()

        for album in artist.albums:
            _artist['albums'].append(albums.get_full_tree(album))

        return _artist

    def delete(self, artist_id=None):
        if not artist_id:
            return JSONResponse(405)

        artist = Artist.query.get(artist_id)
        if not artist:
            abort(404)

        g.db.session.delete(artist)
        g.db.session.commit()

        return {}


class AlbumResource(Resource):
    """ The resource resposible for albums. """

    def get_resource_fields(self):
        return {
            'id': fields.Integer(attribute='pk'),
            'name': fields.String,
            'slug': fields.String,
            'year': fields.Integer,
            'uri': InstanceURI('album'),
            'artists': ManyToManyField(Artist, {
                'id': fields.Integer(attribute='pk'),
                'uri': InstanceURI('artist'),
            }),
            'cover': fields.String(default=app.config['DEFAULT_ALBUM_COVER']),
        }

    def get(self, album_id=None, album_slug=None):
        if not album_id and not album_slug:
            return list(self.get_many())

        if not album_id and album_slug:
            album = self.get_by_slug(album_slug)
        else:
            album = self.get_one(album_id)

        if full_tree():
            return self.get_full_tree(album)

        return marshal(album, self.get_resource_fields())

    def get_many(self):
        artist_pk = request.args.get('artist')
        if artist_pk:
            albums = Album.query.join(Album.artists).filter(
                Artist.pk == artist_pk)
        else:
            albums = Album.query

        queryset = albums.order_by(Album.year, Album.name, Album.pk)
        for album in paginate(queryset):
            yield marshal(album, self.get_resource_fields())

    def get_one(self, album_id):
        album = Album.query.get(album_id)

        if not album:
            abort(404)

        return album

    def get_by_slug(self, album_slug):
        album = Album.query.filter_by(slug=album_slug).first()

        if not album:
            abort(404)

        return album

    def get_full_tree(self, album):
        _album = marshal(album, self.get_resource_fields())
        _album['tracks'] = []

        tracks = TrackResource()

        for track in album.tracks.order_by('number', 'title'):
            _album['tracks'].append(tracks.get_full_tree(track))

        return _album

    def delete(self, album_id=None):
        if not album_id:
            return JSONResponse(405)

        album = Album.query.get(album_id)
        if not album:
            abort(404)

        g.db.session.delete(album)
        g.db.session.commit()

        return {}


class TrackResource(Resource):
    """ The resource resposible for tracks. """

    def get_resource_fields(self):
        return {
            'id': fields.Integer(attribute='pk'),
            'uri': InstanceURI('track'),
            'files': TrackFiles,
            'bitrate': fields.Integer,
            'length': fields.Integer,
            'title': fields.String,
            'slug': fields.String,
            'artist': ForeignKeyField(Artist, {
                'id': fields.Integer(attribute='pk'),
                'uri': InstanceURI('artist'),
            }),
            'album': ForeignKeyField(Album, {
                'id': fields.Integer(attribute='pk'),
                'uri': InstanceURI('album'),
            }),
            'number': fields.Integer,
        }

    def get(self, track_id=None, track_slug=None):
        if not track_id and not track_slug:
            return list(self.get_many())

        if not track_id and track_slug:
            track = self.get_by_slug(track_slug)
        else:
            track = self.get_one(track_id)

        if full_tree():
            return self.get_full_tree(track, include_scraped=True,
                                      include_related=True)

        return marshal(track, self.get_resource_fields())

    # TODO: Pagination
    def get_many(self):
        album_pk = request.args.get('album')
        artist_pk = request.args.get('artist')
        if album_pk:
            album_pk = None if album_pk == 'null' else album_pk
            tracks = Track.query.filter_by(album_pk=album_pk)
        elif artist_pk:
            tracks = Track.query.filter(Track.artist_pk == artist_pk)
        else:
            tracks = Track.query

        queryset = tracks.order_by(Track.album_pk, Track.number, Track.pk)
        for track in paginate(queryset):
            if full_tree():
                yield self.get_full_tree(track, include_related=True)
            else:
                yield marshal(track, self.get_resource_fields())

    def get_one(self, track_id):
        track = Track.query.get(track_id)

        if not track:
            abort(404)

        return track

    def get_by_slug(self, track_slug):
        track = Track.query.filter_by(slug=track_slug).first()

        if not track:
            abort(404)

        return track

    def get_full_tree(self, track, include_scraped=False,
                      include_related=False):
        """
        Retrives the full tree for a track. If the include_related option is
        not set then a normal track structure will be retrieved. If its set
        external resources that need to be scraped, like lyrics, will also be
        included. Also related objects like artist and album will be expanded
        to provide all their respective information.

        This is disabled by default to avois DoS'ing lyrics' websites when
        requesting many tracks at once.

        """

        resource_fields = self.get_resource_fields()
        if include_related:
            artist = ArtistResource()
            resource_fields['artist'] = ForeignKeyField(
                Artist,
                artist.get_resource_fields())
            album = AlbumResource()
            resource_fields['album'] = ForeignKeyField(
                Album,
                album.get_resource_fields())

        _track = marshal(track, resource_fields)

        if include_scraped:
            lyrics = LyricsResource()
            try:
                _track['lyrics'] = lyrics.get_for(track)
            except NotFound:
                _track['lyrics'] = None

        # tabs = TabsResource()
        # _track['tabs'] = tabs.get()

        return _track

    def delete(self, track_id=None):
        if not track_id:
            return JSONResponse(405)

        track = Track.query.get(track_id)
        if not track:
            abort(404)

        g.db.session.delete(track)
        g.db.session.commit()

        return {}


class LyricsResource(Resource):
    """
    The resource responsible for a track's lyrics. Lyrics are scraped on
    demand, and only the URI where they are found is stored in the database.

    """

    def get_resource_fields(self):
        return {
            'id': fields.Integer(attribute='pk'),
            'uri': InstanceURI('lyrics'),
            'text': fields.String,
            'source_uri': fields.String(attribute='source'),
            'track': ForeignKeyField(Track, {
                'id': fields.Integer(attribute='pk'),
                'uri': InstanceURI('track'),
            }),
        }

    def get(self, track_id=None, track_slug=None):
        if not track_id and not track_slug:
            abort(404)

        if not track_id and track_slug:
            track = Track.query.filter_by(slug=track_slug).first()
        else:
            track = Track.query.get(track_id)

        return self.get_for(track)

    def get_for(self, track):
        if track.lyrics:
            return marshal(track.lyrics, self.get_resource_fields())

        try:
            lyrics = get_lyrics(track)
        except:
            log.debug(traceback.format_exc())
            lyrics = None

        if not lyrics:
            abort(404)

        return marshal(lyrics, self.get_resource_fields())

    def post(self, track_id):
        text = request.form.get('text', None)
        if not text:
            return JSONResponse(400)

        track = Track.query.get(track_id)
        lyric = LyricsCache(track=track, text=text)

        g.db.session.add(lyric)
        g.db.commit()

        return JSONResponse(200)

    def delete(self, track_id):
        track = Track.query.get(track_id)
        g.db.session.delete(track.lyrics)
        g.db.session.commit()

        return JSONResponse(200)


class ConvertResource(Resource):
    """ Resource in charge of converting tracks from one format to another. """

    def get(self, track_id):
        track = Track.query.get(track_id)
        mimetype = request.args.get('mimetype')
        if not track or not mimetype:
            abort(404)

        ConverterClass = get_converter()
        try:
            converter = ConverterClass(track, mimetype=mimetype)
        except InvalidMimeTypeError, e:
            log.error(e)
            abort(400)

        converter.convert()
        uri = converter.get_uri()

        return JSONResponse(status=301, headers={'Location': uri})


class ShowsResource(Resource):
    """
    """

    def get_resource_fields(self):
        return {
            'id': fields.String,
            'artists': ManyToManyField(Artist, {
                'id': fields.Integer(attribute='pk'),
                'uri': InstanceURI('artist'),
            }),
            'other_artists': fields.List(fields.Raw),
            'datetime': fields.DateTime,
            'title': fields.String,
            'tickets_left': Boolean,
            'venue': fields.Nested({
                'latitude': fields.String,
                'longitude': fields.String,
                'name': fields.String,
            }),
        }

    def get(self, artist_id=None, artist_slug=None):
        if not artist_id and not artist_slug:
            abort(404)

        if not artist_id and artist_slug:
            artist = Artist.query.filter_by(slug=artist_slug).first()
        else:
            artist = Artist.query.get(artist_id)

        if not artist:
            abort(404)

        latitude = request.args.get('latitude')
        longitude = request.args.get('longitude')

        country = request.args.get('country')
        city = request.args.get('city')

        if latitude and longitude:
            location = (latitude, longitude)
        elif country and city:
            location = (city, country)
        else:
            location = ()

        response = self.fetch(artist.name, location)

        return list(response) if response else []

    def fetch(self, artist_name, location):
        bit_uri = ('http://api.bandsintown.com/artists/%(artist)s/events'
                   '/search?format=json&app_id=%(app_id)s&api_version=2.0')
        bit_uri = bit_uri % {
            'artist': urllib2.quote(artist_name),
            'app_id': app.config['BANDSINTOWN_APP_ID'],
        }

        _location = urllib2.quote('%s, %s' % location) if location else \
            'use_geoip'

        bit_uri = '&'.join((bit_uri, '='.join(('location', _location))))

        log.info(bit_uri)

        try:
            response = requests.get(bit_uri)
        except requests.exceptions.RequestException:
            return

        for event in response.json():
            yield marshal(ShowModel(artist_name, event),
                          self.get_resource_fields())


class RandomResource(Resource):
    """ Retrieves a random instance of a specified resource. """

    def get(self, resource_name):
        get_resource = getattr(self, 'get_%s' % resource_name)

        if get_resource and callable(get_resource):
            resource_fields = {
                'id': fields.Integer(attribute='pk'),
                'uri': InstanceURI(resource_name),
            }

            return marshal(get_resource(), resource_fields)

        abort(404)

    def get_track(self):
        return Track.random()

    def get_album(self):
        return Album.random()

    def get_artist(self):
        return Artist.random()


class WhatsNewResource(Resource):
    """
    Resource consisting of artists, albums and tracks that were indexed after a
    given date.

    """

    def get(self):
        news = {'artists': [], 'albums': [], 'tracks': []}
        try:
            self.since = datetime.strptime(request.args.get('since'), '%Y%m%d')
        except:
            log.error(traceback.format_exc())
            return news

        news = {
            'artists': self.get_new_for(Artist, 'artist'),
            'albums': self.get_new_for(Album, 'album'),
            'tracks': self.get_new_for(Track, 'track'),
        }

        return news

    def get_new_for(self, model, resource_name):
        """
        Fetches all the instances with ``date_added`` older than
        ``self.since``.

        """

        query = model.query.filter(model.date_added > self.since)
        resource_fields = {
            'id': fields.Integer(attribute='pk'),
            'uri': InstanceURI(resource_name),
        }

        return [marshal(row, resource_fields) for row in query.all()]


class ClientResource(Resource):
    """ Resource that lists the known clients for Shiva. """

    def get(self):
        clients = [
            {
                'name': 'Shiva-Client',
                'uri': 'https://github.com/tooxie/shiva-client',
                'authors': [
                    u'Alvaro Mouriño <alvaro@mourino.net>',
                ],
            },
            {
                'name': 'Shiva4J',
                'uri': 'https://github.com/instant-solutions/shiva4j',
                'authors': [
                    u'instant:solutions <office@instant-it.at>'
                ],
            },
            {
                'name': 'Shakti',
                'uri': 'https://github.com/gonz/shakti',
                'authors': [
                    u'Gonzalo Saavedra <gonzalosaavedra@gmail.com>',
                ],
            },
        ]

        return clients


class AboutResource(Resource):
    """ Just some information about Shiva. """

    def get(self):
        info = {
            'name': 'Shiva',
            'version': get_version(),
            'author': u'Alvaro Mouriño <alvaro@mourino.net>',
            'uri': 'https://github.com/tooxie/shiva-server',
            'contributors': get_contributors(),
        }

        return info



import re
import urllib2
import urllib

import lxml.html
import requests
from flask import current_app as app

from shiva.lyrics import LyricScraper
from shiva.utils import get_logger

log = get_logger()


class MetroLyrics(LyricScraper):
    """
    """

    def __init__(self, artist, title):
        self.artist = artist
        self.title = title
        self.search_url = 'http://www.metrolyrics.com/api/v1/search/artistsong'
        self.title_re = re.compile(r'<title>(?P<artist>.*?) - '
                                   r'(?P<title>.*?) LYRICS</title>')

        self.lyrics = None
        self.source = None

    def fetch(self):
        self.search()
        if not self.source:
            return None

        response = requests.get(self.source)
        self.html = response.text
        html = lxml.html.fromstring(self.html)

        if not self.check():
            return False

        log.info('[FOUND] %s' % self.source)
        div = html.get_element_by_id('lyrics-body')
        lyrics = re.sub(r'\n\[ From: .*? \]', '', div.text_content())

        self.lyrics = lyrics.strip()

        return True

    def search(self):
        params = {'artist': self.artist,
                  'song': self.title,
                  'X-API-KEY': app.config['METROLYRICS_API_KEY']}
        _url = '?'.join((self.search_url, urllib.urlencode(params)))
        log.info('[SEARCH] %s' % _url)
        response = requests.get(_url)
        if response.status_code == 200:
            self.source = response.json()['items'][0]['url']

    def check(self):
        match = self.title_re.search(self.html)

        if match.group('artist').lower() != self.artist.lower():
            return False

        if match.group('title').lower() != self.title.lower():
            return False

        return True
